// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use hashbrown::{hash_map::Entry, HashMap};
use petgraph::{
    graph::{NodeIndex, UnGraph},
    unionfind::UnionFind,
};

use crate::{
    bundle::{BundleField, BundleType},
    enum_::{EnumType, EnumVariant},
    expr::{ops, ExprEnum, Flow},
    intern::{Intern, Interned, Memoize},
    module::{
        transform::visit::{Visit, Visitor},
        AnnotatedModuleIO, Block, ExternModuleBody, InstantiatedModule, ModuleBody, ModuleIO,
        NormalModuleBody, Stmt, StmtConnect, StmtDeclaration, StmtFormal, StmtIf, StmtInstance,
        StmtMatch, StmtReg, StmtWire, TargetInInstantiatedModule,
    },
    prelude::*,
};
use std::{convert::Infallible, fmt, mem};

#[derive(Debug)]
pub enum DeduceResetsError {
    ResetIsNotDrivenByAsyncOrSync { target: TargetInInstantiatedModule },
}

impl fmt::Display for DeduceResetsError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            DeduceResetsError::ResetIsNotDrivenByAsyncOrSync { target } => {
                write!(f, "deduce_reset failed: Reset signal is not driven by any AsyncReset or SyncReset signals: {target:?}")
            }
        }
    }
}

impl std::error::Error for DeduceResetsError {}

impl From<DeduceResetsError> for std::io::Error {
    fn from(value: DeduceResetsError) -> Self {
        std::io::Error::new(std::io::ErrorKind::Other, value)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
enum ResetTarget {
    Base {
        base: Expr<CanonicalType>,
    },
    BundleField {
        parent: Interned<ResetTarget>,
        field_ty: CanonicalType,
        field_index: usize,
    },
    EnumVariant {
        parent: Interned<ResetTarget>,
        variant_ty: CanonicalType,
        variant_index: usize,
    },
    /// Array's Elements:
    /// deduce_resets requires all array elements to deduce to the same pattern of async/sync resets,
    /// so we don't track individual array elements but instead track all of an array's elements together.
    ArraysElements {
        parent: Interned<ResetTarget>,
        element_ty: CanonicalType,
    },
}

impl ResetTarget {
    fn canonical_ty(self) -> CanonicalType {
        match self {
            ResetTarget::Base { base } => Expr::ty(base),
            ResetTarget::BundleField { field_ty, .. } => field_ty,
            ResetTarget::EnumVariant { variant_ty, .. } => variant_ty,
            ResetTarget::ArraysElements { element_ty, .. } => element_ty,
        }
    }
    fn parent(self) -> Option<Interned<ResetTarget>> {
        match self {
            ResetTarget::Base { .. } => None,
            ResetTarget::BundleField { parent, .. }
            | ResetTarget::EnumVariant { parent, .. }
            | ResetTarget::ArraysElements { parent, .. } => Some(parent),
        }
    }
    fn base(mut self) -> Expr<CanonicalType> {
        loop {
            match self {
                ResetTarget::Base { base } => break base,
                ResetTarget::BundleField { parent, .. }
                | ResetTarget::EnumVariant { parent, .. }
                | ResetTarget::ArraysElements { parent, .. } => self = *parent,
            }
        }
    }
    fn bundle_field(self, field_index: usize) -> Self {
        let field_ty = Bundle::from_canonical(self.canonical_ty()).fields()[field_index].ty;
        Self::BundleField {
            parent: self.intern_sized(),
            field_ty,
            field_index,
        }
    }
    fn enum_variant(self, variant_index: usize) -> Self {
        let variant_ty = Enum::from_canonical(self.canonical_ty()).variants()[variant_index]
            .ty
            .expect("known to have a variant field");
        Self::EnumVariant {
            parent: self.intern_sized(),
            variant_ty,
            variant_index,
        }
    }
    fn arrays_elements(self) -> Self {
        let element_ty = <Array>::from_canonical(self.canonical_ty()).element();
        Self::ArraysElements {
            parent: self.intern_sized(),
            element_ty,
        }
    }
    fn for_each_child<E>(self, mut f: impl FnMut(ResetTarget) -> Result<(), E>) -> Result<(), E> {
        match self.canonical_ty() {
            CanonicalType::UInt(_)
            | CanonicalType::SInt(_)
            | CanonicalType::Bool(_)
            | CanonicalType::AsyncReset(_)
            | CanonicalType::SyncReset(_)
            | CanonicalType::Reset(_)
            | CanonicalType::Clock(_) => Ok(()),
            CanonicalType::Array(_) => f(self.arrays_elements()),
            CanonicalType::Enum(ty) => {
                for variant_index in 0..ty.variants().len() {
                    f(self.enum_variant(variant_index))?;
                }
                Ok(())
            }
            CanonicalType::Bundle(ty) => {
                for field_index in 0..ty.fields().len() {
                    f(self.bundle_field(field_index))?;
                }
                Ok(())
            }
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
struct ResetTargetInInstantiatedModule {
    instantiated_module: InstantiatedModule,
    target: ResetTarget,
}

#[derive(Debug)]
struct Node {
    target: ResetTargetInInstantiatedModule,
    deduced_type: Option<CanonicalType>,
}

#[derive(Debug)]
struct State {
    base_module: Interned<Module<Bundle>>,
    node_ids: HashMap<ResetTargetInInstantiatedModule, NodeIndex<u32>>,
    graph: UnGraph<Node, ()>,
    fallback_to_sync_reset: bool,
}

fn type_contains_any_undeduced_resets(ty: CanonicalType) -> bool {
    #[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
    struct MyMemoize;
    impl Memoize for MyMemoize {
        type Input = CanonicalType;
        type InputOwned = CanonicalType;
        type Output = bool;

        fn inner(self, input: &Self::Input) -> Self::Output {
            match input {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => false,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(ty) => type_contains_any_undeduced_resets(ty.element()),
                CanonicalType::Enum(ty) => {
                    ty.variants().iter().any(|EnumVariant { ty, .. }| {
                        ty.is_some_and(|ty| type_contains_any_undeduced_resets(ty))
                    })
                }
                CanonicalType::Bundle(ty) => ty
                    .fields()
                    .iter()
                    .any(|&BundleField { ty, .. }| type_contains_any_undeduced_resets(ty)),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => true,
                CanonicalType::Clock(_) => false,
            }
        }
    }
    MyMemoize.get_owned(ty)
}

trait ProcessStep {
    type Error;
    type Processed<T>;
    fn processed_as_ref<T>(v: &Self::Processed<T>) -> Self::Processed<&T>;
    fn processed_as_mut<T>(v: &mut Self::Processed<T>) -> Self::Processed<&mut T>;
    fn processed_zip<T, U>(t: Self::Processed<T>, u: Self::Processed<U>)
        -> Self::Processed<(T, U)>;
    fn processed_map<T, U>(v: Self::Processed<T>, f: impl FnOnce(T) -> U) -> Self::Processed<U>;
    fn processed_make<T>(t: T) -> Self::Processed<T>;
    fn processed_from_iter<T, O: FromIterator<T>>(
        iter: impl IntoIterator<Item = Self::Processed<T>>,
    ) -> Self::Processed<O>;
}

macro_rules! impl_process_step_with_empty_processed {
    () => {
        type Processed<T> = ();
        fn processed_as_ref<T>(v: &Self::Processed<T>) -> Self::Processed<&T> {
            *v
        }
        fn processed_as_mut<T>(v: &mut Self::Processed<T>) -> Self::Processed<&mut T> {
            *v
        }
        fn processed_zip<T, U>(
            t: Self::Processed<T>,
            u: Self::Processed<U>,
        ) -> Self::Processed<(T, U)> {
            let _ = t;
            let _ = u;
            ()
        }
        fn processed_map<T, U>(
            v: Self::Processed<T>,
            f: impl FnOnce(T) -> U,
        ) -> Self::Processed<U> {
            let _ = f;
            v
        }
        fn processed_make<T>(t: T) -> Self::Processed<T> {
            let _ = t;
            ()
        }
        fn processed_from_iter<T, O: FromIterator<T>>(
            iter: impl IntoIterator<Item = Self::Processed<T>>,
        ) -> Self::Processed<O> {
            FromIterator::from_iter(iter)
        }
    };
}

struct Processed<T, Step: ProcessStep>(Step::Processed<T>);

impl<T, Step: ProcessStep> Processed<T, Step> {
    fn as_ref(&self) -> Processed<&T, Step> {
        Processed(Step::processed_as_ref(&self.0))
    }
    fn as_mut(&mut self) -> Processed<&mut T, Step> {
        Processed(Step::processed_as_mut(&mut self.0))
    }
    fn zip<U>(self, u: Processed<U, Step>) -> Processed<(T, U), Step> {
        Processed(Step::processed_zip(self.0, u.0))
    }
    fn new(v: T) -> Self {
        Processed(Step::processed_make(v))
    }
    fn map<U>(self, f: impl FnOnce(T) -> U) -> Processed<U, Step> {
        Processed(Step::processed_map(self.0, f))
    }
}

impl<C: FromIterator<A>, A, Step: ProcessStep> FromIterator<Processed<A, Step>>
    for Processed<C, Step>
{
    fn from_iter<T: IntoIterator<Item = Processed<A, Step>>>(iter: T) -> Self {
        Processed(Step::processed_from_iter(iter.into_iter().map(|v| v.0)))
    }
}

fn for_each_subexpr_exclusive<E>(
    expr: Expr<CanonicalType>,
    f: impl FnMut(Expr<CanonicalType>) -> Result<(), E>,
) -> Result<(), E> {
    struct MyVisitor<F> {
        f: F,
    }
    impl<F: FnMut(Expr<CanonicalType>) -> Result<(), E>, E> Visitor for MyVisitor<F> {
        type Error = E;
        fn visit_expr_enum(&mut self, v: &ExprEnum) -> Result<(), Self::Error> {
            (self.f)(v.to_expr())?;
            v.default_visit(self)
        }
    }
    Expr::expr_enum(expr).default_visit(&mut MyVisitor { f })
}

struct AddNodesToGraphStep;

impl ProcessStep for AddNodesToGraphStep {
    type Error = Infallible;
    impl_process_step_with_empty_processed!();
}

struct AddEdgesToGraphStep {
    union_find: UnionFind<u32>,
}

impl ProcessStep for AddEdgesToGraphStep {
    type Error = Infallible;
    impl_process_step_with_empty_processed!();
}

trait RunProcessStep<Step: ProcessStep>: Sized {
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error>;
}

impl RunProcessStep<AddNodesToGraphStep> for ResetTarget {
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut AddNodesToGraphStep,
    ) -> Result<Processed<Self, AddNodesToGraphStep>, Infallible> {
        let target = ResetTargetInInstantiatedModule {
            instantiated_module,
            target: self,
        };
        let Entry::Vacant(entry) = state.node_ids.entry(target) else {
            return Ok(Processed(()));
        };
        let ty = self.canonical_ty();
        let node_id = state.graph.add_node(Node {
            target,
            deduced_type: type_contains_any_undeduced_resets(ty).then_some(ty),
        });
        entry.insert(node_id);
        self.for_each_child(|target| -> Result<(), Infallible> {
            target.run_process_step(instantiated_module, state, step)?;
            Ok(())
        })?;
        Ok(Processed(()))
    }
}

impl RunProcessStep<AddEdgesToGraphStep> for ResetTarget {
    fn run_process_step(
        self,
        _instantiated_module: InstantiatedModule,
        _state: &mut State,
        _step: &mut AddEdgesToGraphStep,
    ) -> Result<Processed<Self, AddEdgesToGraphStep>, Infallible> {
        Ok(Processed(()))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for Expr<CanonicalType>
where
    ResetTarget: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let retval = ResetTarget::Base { base: self }
            .run_process_step(instantiated_module, state, step)?
            .map(|target| match target {
                ResetTarget::Base { base } => base,
                _ => unreachable!(),
            });
        for_each_subexpr_exclusive(self, |base| {
            ResetTarget::Base { base }.run_process_step(instantiated_module, state, step)?;
            Ok(())
        })?;
        Ok(retval)
    }
}

struct ConnectAndLhsInstantiatedModule {
    lhs_instantiated_module: InstantiatedModule,
    lhs: Expr<CanonicalType>,
    rhs: Expr<CanonicalType>,
    source_location: SourceLocation,
}

impl RunProcessStep<AddNodesToGraphStep> for ConnectAndLhsInstantiatedModule {
    fn run_process_step(
        self,
        _instantiated_module: InstantiatedModule,
        _state: &mut State,
        _step: &mut AddNodesToGraphStep,
    ) -> Result<Processed<Self, AddNodesToGraphStep>, Infallible> {
        Ok(Processed(()))
    }
}

impl RunProcessStep<AddEdgesToGraphStep> for ConnectAndLhsInstantiatedModule {
    fn run_process_step(
        self,
        rhs_instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut AddEdgesToGraphStep,
    ) -> Result<Processed<Self, AddEdgesToGraphStep>, Infallible> {
        let Self {
            lhs_instantiated_module,
            lhs,
            rhs,
            source_location,
        } = self;
        match Expr::ty(lhs) {
            CanonicalType::UInt(_) => {}
            CanonicalType::SInt(_) => {}
            CanonicalType::Bool(_) => {}
            CanonicalType::AsyncReset(ty) => todo!(),
            CanonicalType::SyncReset(ty) => todo!(),
            CanonicalType::Reset(ty) => todo!(),
            CanonicalType::Clock(_) => {}
            CanonicalType::Array(ty) => todo!(),
            CanonicalType::Enum(ty) => {
                for (variant_index, EnumVariant { ty, .. }) in ty.variants().into_iter().enumerate()
                {
                    if ty.is_none() {
                        continue;
                    }
                    let lhs =
                        ops::VariantAccess::new_by_index(Expr::from_canonical(lhs), variant_index)
                            .to_expr();
                    let rhs =
                        ops::VariantAccess::new_by_index(Expr::from_canonical(rhs), variant_index)
                            .to_expr();
                    ConnectAndLhsInstantiatedModule {
                        lhs_instantiated_module,
                        lhs,
                        rhs,
                        source_location,
                    }
                    .run_process_step(rhs_instantiated_module, state, step)?;
                }
            }
            CanonicalType::Bundle(ty) => {
                for (field_index, BundleField { flipped, .. }) in
                    ty.fields().into_iter().enumerate()
                {
                    let mut lhs =
                        ops::FieldAccess::new_by_index(Expr::from_canonical(lhs), field_index)
                            .to_expr();
                    let mut rhs =
                        ops::FieldAccess::new_by_index(Expr::from_canonical(rhs), field_index)
                            .to_expr();
                    let mut lhs_instantiated_module = lhs_instantiated_module;
                    let mut rhs_instantiated_module = rhs_instantiated_module;
                    if flipped {
                        mem::swap(&mut lhs, &mut rhs);
                        mem::swap(&mut lhs_instantiated_module, &mut rhs_instantiated_module);
                    }
                    ConnectAndLhsInstantiatedModule {
                        lhs_instantiated_module,
                        lhs,
                        rhs,
                        source_location,
                    }
                    .run_process_step(rhs_instantiated_module, state, step)?;
                }
            }
        }
        Ok(Processed(()))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtConnect
where
    ConnectAndLhsInstantiatedModule: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let StmtConnect {
            lhs,
            rhs,
            source_location,
        } = self;
        Ok(ConnectAndLhsInstantiatedModule {
            lhs_instantiated_module: instantiated_module,
            lhs,
            rhs,
            source_location,
        }
        .run_process_step(instantiated_module, state, step)?
        .map(
            |ConnectAndLhsInstantiatedModule {
                 lhs_instantiated_module: _,
                 lhs,
                 rhs,
                 source_location,
             }| StmtConnect {
                lhs,
                rhs,
                source_location,
            },
        ))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtFormal {
    fn run_process_step(
        self,
        _instantiated_module: InstantiatedModule,
        _state: &mut State,
        _step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        // no inputs are Reset
        Ok(Processed::new(self))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtIf {
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        todo!()
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtMatch {
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        todo!()
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for ModuleIO<CanonicalType>
where
    ResetTarget: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        Ok(self
            .to_expr()
            .run_process_step(instantiated_module, state, step)?
            .map(|expr| match *Expr::expr_enum(expr) {
                ExprEnum::ModuleIO(module_io) => module_io,
                _ => unreachable!(),
            }))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtWire
where
    ResetTarget: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let Self { annotations, wire } = self;
        Ok(wire
            .to_expr()
            .run_process_step(instantiated_module, state, step)?
            .map(|expr| match *Expr::expr_enum(expr) {
                ExprEnum::Wire(wire) => Self { annotations, wire },
                _ => unreachable!(),
            }))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtReg
where
    ResetTarget: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let Self { annotations, reg } = self;
        Ok(reg
            .to_expr()
            .run_process_step(instantiated_module, state, step)?
            .map(|expr| match *Expr::expr_enum(expr) {
                ExprEnum::Reg(reg) => Self { annotations, reg },
                _ => unreachable!(),
            }))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtInstance
where
    ResetTarget: RunProcessStep<Step>,
    ConnectAndLhsInstantiatedModule: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let Self {
            annotations,
            instance,
        } = self;
        let child_instantiated_module = InstantiatedModule::Child {
            parent: instantiated_module.intern_sized(),
            instance: instance.intern_sized(),
        };
        instance
            .instantiated()
            .run_process_step(child_instantiated_module, state, step)?;
        for (field_index, AnnotatedModuleIO { module_io, .. }) in
            instance.instantiated().module_io().into_iter().enumerate()
        {
            let (lhs_instantiated_module, lhs, rhs_instantiated_module, rhs) = match module_io
                .flow()
            {
                Flow::Source => {
                    // connect to submodule's input from instance input
                    (
                        child_instantiated_module,
                        module_io.to_expr(),
                        instantiated_module,
                        ops::FieldAccess::new_by_index(instance.to_expr(), field_index).to_expr(),
                    )
                }
                Flow::Sink => {
                    // connect to instance output from submodule's output
                    (
                        instantiated_module,
                        ops::FieldAccess::new_by_index(instance.to_expr(), field_index).to_expr(),
                        child_instantiated_module,
                        module_io.to_expr(),
                    )
                }
                Flow::Duplex => unreachable!(),
            };
            ConnectAndLhsInstantiatedModule {
                lhs_instantiated_module,
                lhs,
                rhs,
                source_location: instance.source_location(),
            }
            .run_process_step(rhs_instantiated_module, state, step)?;
        }
        Ok(Expr::canonical(instance.to_expr())
            .run_process_step(instantiated_module, state, step)?
            .map(|expr| match *Expr::expr_enum(expr) {
                ExprEnum::Instance(instance) => Self {
                    annotations,
                    instance,
                },
                _ => unreachable!(),
            }))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for StmtDeclaration
where
    ResetTarget: RunProcessStep<Step>,
    ConnectAndLhsInstantiatedModule: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        Ok(match self {
            StmtDeclaration::Wire(decl) => decl
                .run_process_step(instantiated_module, state, step)?
                .map(StmtDeclaration::from),
            StmtDeclaration::Reg(decl) => decl
                .run_process_step(instantiated_module, state, step)?
                .map(StmtDeclaration::from),
            StmtDeclaration::Instance(decl) => decl
                .run_process_step(instantiated_module, state, step)?
                .map(StmtDeclaration::from),
        })
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for Stmt
where
    ResetTarget: RunProcessStep<Step>,
    ConnectAndLhsInstantiatedModule: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        Ok(match self {
            Stmt::Connect(stmt) => stmt
                .run_process_step(instantiated_module, state, step)?
                .map(Stmt::from),
            Stmt::Formal(stmt) => stmt
                .run_process_step(instantiated_module, state, step)?
                .map(Stmt::from),
            Stmt::If(stmt) => stmt
                .run_process_step(instantiated_module, state, step)?
                .map(Stmt::from),
            Stmt::Match(stmt) => stmt
                .run_process_step(instantiated_module, state, step)?
                .map(Stmt::from),
            Stmt::Declaration(stmt) => stmt
                .run_process_step(instantiated_module, state, step)?
                .map(Stmt::from),
        })
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for Block
where
    ResetTarget: RunProcessStep<Step>,
    ConnectAndLhsInstantiatedModule: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let Block { memories, stmts } = self;
        // memories and memory ports won't ever contain any Reset values,
        // so always just use the old `memories` value. we add the ports to the graph anyway to make the other logic easier.
        for memory in memories {
            for port in memory.ports() {
                Expr::canonical(port.to_expr()).run_process_step(
                    instantiated_module,
                    state,
                    step,
                )?;
            }
        }
        let stmts = Result::<Processed<_, _>, _>::from_iter(
            stmts
                .iter()
                .map(|stmt| stmt.run_process_step(instantiated_module, state, step)),
        )?;
        Ok(stmts.map(|stmts| Block { memories, stmts }))
    }
}

impl<Step: ProcessStep> RunProcessStep<Step> for Module<Bundle>
where
    ResetTarget: RunProcessStep<Step>,
    ConnectAndLhsInstantiatedModule: RunProcessStep<Step>,
{
    fn run_process_step(
        self,
        instantiated_module: InstantiatedModule,
        state: &mut State,
        step: &mut Step,
    ) -> Result<Processed<Self, Step>, Step::Error> {
        let module = *instantiated_module.leaf_module();
        let module_io =
            Result::<Processed<Interned<[_]>, _>, _>::from_iter(module.module_io().iter().map(
                |&AnnotatedModuleIO {
                     annotations,
                     module_io,
                 }| {
                    Ok(module_io
                        .run_process_step(instantiated_module, state, step)?
                        .map(|module_io| AnnotatedModuleIO {
                            annotations,
                            module_io,
                        }))
                },
            ))?;
        let body = match module.body() {
            ModuleBody::Normal(NormalModuleBody { body }) => body
                .run_process_step(instantiated_module, state, step)?
                .map(|body| ModuleBody::Normal(NormalModuleBody { body })),
            body @ ModuleBody::Extern(ExternModuleBody {
                verilog_name: _,
                parameters: _,
            }) => Processed::new(body),
        };
        Ok(module_io.zip(body).map(|(module_io, body)| {
            Module::new_unchecked(
                module.name_id(),
                module.source_location(),
                body,
                module_io,
                module.module_annotations(),
            )
        }))
    }
}

pub fn deduce_resets(
    module: Interned<Module<Bundle>>,
    fallback_to_sync_reset: bool,
) -> Result<Interned<Module<Bundle>>, DeduceResetsError> {
    let mut state = State {
        base_module: module,
        node_ids: HashMap::new(),
        graph: UnGraph::new_undirected(),
        fallback_to_sync_reset,
    };
    let Ok(Processed(())) = module.run_process_step(
        InstantiatedModule::Base(module),
        &mut state,
        &mut AddNodesToGraphStep,
    );
    let mut step = AddEdgesToGraphStep {
        union_find: UnionFind::new(state.graph.node_count()),
    };
    let Ok(Processed(())) =
        module.run_process_step(InstantiatedModule::Base(module), &mut state, &mut step);
    let AddEdgesToGraphStep { union_find } = step;
    todo!("deduce types");
    Ok(todo!("transform module"))
}
